@@ -2,6 +2,7 @@
 
 import os
 from ctypes import (
+    CFUNCTYPE,
     POINTER,
     Structure,
     _Pointer,
@@ -11,20 +12,32 @@ from ctypes import (
     c_int,
     c_size_t,
     c_uint32,
+    c_void_p,
     cdll,
     create_string_buffer,
+    string_at,
 )
 from ctypes.util import find_library
+from dataclasses import dataclass
 from enum import Enum, IntFlag
-from functools import reduce
+from functools import partial, reduce
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, NamedTuple, Optional, TypeAlias
+from typing import TYPE_CHECKING, Iterable, NamedTuple, Optional, TypeAlias
 
 ###############################################################################
 # Types
 ###############################################################################
 
 
+class allocated_c_char_p(c_char_p):
+    """
+    This class is used in place of c_char_p when we need to free it manually.
+    Python would convert c_char_p to a bytes string and we would not be able to free it.
+    """
+
+    pass
+
+
 class xkb_context(Structure):
     pass
 
@@ -60,10 +73,10 @@ if TYPE_CHECKING:
     xkb_keymap_p: TypeAlias = _Pointer[xkb_keymap]
     xkb_state_p = _Pointer[xkb_state]
 else:
-    xkb_context_p = Any
-    xkb_rule_names_p = Any
-    xkb_keymap_p = Any
-    xkb_state_p = Any
+    xkb_context_p = POINTER(xkb_context)
+    xkb_rule_names_p = POINTER(xkb_rule_names)
+    xkb_keymap_p = POINTER(xkb_keymap)
+    xkb_state_p = POINTER(xkb_state)
 xkb_context_flags = c_int
 xkb_keymap_compile_flags = c_int
 xkb_keymap_format = c_int
@@ -75,6 +88,9 @@ xkb_level_index_t = c_uint32
 xkb_layout_index_t = c_uint32
 xkb_state_component = c_int
 xkb_consumed_mode = c_int
+xkb_keysym_flags = c_int
+
+xkb_keymap_key_iter_t = CFUNCTYPE(None, xkb_keymap_p, xkb_keycode_t, c_void_p)
 
 
 ###############################################################################
@@ -85,9 +101,11 @@ xkb_consumed_mode = c_int
 XKB_CONTEXT_NO_DEFAULT_INCLUDES = 1 << 0
 XKB_CONTEXT_NO_ENVIRONMENT_NAMES = 1 << 1
 XKB_KEYCODE_INVALID = 0xFFFFFFFF
+XKB_MOD_INVALID = 0xFFFFFFFF
 XKB_STATE_MODS_EFFECTIVE = 1 << 3
 XKB_CONSUMED_MODE_XKB = 0
 XKB_KEYMAP_FORMAT_TEXT_V1 = 1
+XKB_KEYSYM_NO_FLAGS = 0
 
 
 class ModifierMask(IntFlag):
@@ -131,6 +149,9 @@ else:
     else:
         raise OSError("Cannot load libxbcommon")
 
+xkbcommon.xkb_keysym_from_name.argtypes = [c_char_p, xkb_keysym_flags]
+xkbcommon.xkb_keysym_from_name.restype = xkb_keysym_t
+
 xkbcommon.xkb_context_new.argtypes = [xkb_context_flags]
 xkbcommon.xkb_context_new.restype = POINTER(xkb_context)
 
@@ -144,6 +165,36 @@ xkbcommon.xkb_keymap_new_from_names.restype = POINTER(xkb_keymap)
 xkbcommon.xkb_keymap_key_by_name.argtypes = [POINTER(xkb_keymap), c_char_p]
 xkbcommon.xkb_keymap_key_by_name.restype = xkb_keycode_t
 
+xkbcommon.xkb_keymap_get_as_string.argtypes = [POINTER(xkb_keymap), xkb_keymap_format]
+# Note: should be c_char_p; see comment in the definiton of allocated_c_char_p.
+xkbcommon.xkb_keymap_get_as_string.restype = allocated_c_char_p
+
+xkbcommon.xkb_keymap_key_for_each.argtypes = [
+    xkb_keymap_p,
+    xkb_keymap_key_iter_t,
+    c_void_p,
+]
+xkbcommon.xkb_keymap_key_for_each.restype = None
+
+xkbcommon.xkb_keymap_num_layouts_for_key.argtypes = [xkb_keymap_p, xkb_keycode_t]
+xkbcommon.xkb_keymap_num_layouts_for_key.restype = xkb_layout_index_t
+
+xkbcommon.xkb_keymap_num_levels_for_key.argtypes = [
+    xkb_keymap_p,
+    xkb_keycode_t,
+    xkb_layout_index_t,
+]
+xkbcommon.xkb_keymap_num_levels_for_key.restype = xkb_level_index_t
+
+xkbcommon.xkb_keymap_key_get_syms_by_level.argtypes = [
+    xkb_keymap_p,
+    xkb_keycode_t,
+    xkb_layout_index_t,
+    xkb_level_index_t,
+    POINTER(POINTER(xkb_keysym_t)),
+]
+xkbcommon.xkb_keymap_key_get_syms_by_level.restype = c_int
+
 xkbcommon.xkb_state_new.argtypes = [POINTER(xkb_keymap)]
 xkbcommon.xkb_state_new.restype = POINTER(xkb_state)
 
@@ -169,6 +220,12 @@ xkbcommon.xkb_state_key_get_level.restype = xkb_level_index_t
 xkbcommon.xkb_keymap_num_mods.argtypes = [POINTER(xkb_keymap)]
 xkbcommon.xkb_keymap_num_mods.restype = xkb_mod_index_t
 
+xkbcommon.xkb_keymap_mod_get_name.argtypes = [POINTER(xkb_keymap), xkb_mod_index_t]
+xkbcommon.xkb_keymap_mod_get_name.restype = c_char_p
+
+xkbcommon.xkb_keymap_mod_get_index.argtypes = [POINTER(xkb_keymap), c_char_p]
+xkbcommon.xkb_keymap_mod_get_index.restype = xkb_mod_index_t
+
 xkbcommon.xkb_state_mod_index_is_active.argtypes = [
     POINTER(xkb_state),
     xkb_mod_index_t,
@@ -190,6 +247,11 @@ xkbcommon.xkb_keymap_num_leds.restype = xkb_led_index_t
 xkbcommon.xkb_state_led_index_is_active.argtypes = [POINTER(xkb_state), xkb_led_index_t]
 xkbcommon.xkb_state_led_index_is_active.restype = c_int
 
+if libc_path := find_library("c"):
+    libc = cdll.LoadLibrary(libc_path)
+else:
+    raise ValueError("Cannot import libc")
+
 
 def load_keymap(
     xkb_config_root: Path,
@@ -230,6 +292,38 @@ def unref_keymap(keymap: xkb_keymap_p) -> None:
     xkbcommon.xkb_keymap_unref(keymap)
 
 
+def keymap_as_string(keymap: xkb_keymap_p) -> str:
+    p = xkbcommon.xkb_keymap_get_as_string(keymap, XKB_KEYMAP_FORMAT_TEXT_V1)
+    s = string_at(p).decode("utf-8")
+    libc.free(p)
+    return s
+
+
+def xkb_keymap_key_for_each(keymap: xkb_keymap_p, f, x):
+    xkbcommon.xkb_keymap_key_for_each(keymap, xkb_keymap_key_iter_t(f), x)
+
+
+def xkb_keymap_num_layouts_for_key(keymap: xkb_keymap_p, keycode: int) -> int:
+    return xkbcommon.xkb_keymap_num_layouts_for_key(keymap, keycode)
+
+
+def xkb_keymap_num_levels_for_key(
+    keymap: xkb_keymap_p, keycode: int, layout: int
+) -> int:
+    return xkbcommon.xkb_keymap_num_levels_for_key(keymap, keycode, layout)
+
+
+def xkb_keymap_key_get_syms_by_level(
+    keymap: xkb_keymap_p, keycode: int, layout: int, level: int
+) -> tuple[int, ...]:
+    keysyms_array_t = POINTER(xkb_keysym_t)
+    keysyms_array = keysyms_array_t()
+    count = xkbcommon.xkb_keymap_key_get_syms_by_level(
+        keymap, keycode, layout, level, byref(keysyms_array)
+    )
+    return tuple(keysyms_array[k] for k in range(0, count))
+
+
 def new_state(keymap: xkb_keymap_p) -> xkb_state_p:
     state = xkbcommon.xkb_state_new(keymap)
     if not state:
@@ -253,10 +347,46 @@ def unref_state(state: xkb_state_p) -> None:
     xkbcommon.xkb_state_unref(state)
 
 
+def xkb_keymap_mod_get_index(keymap: xkb_keymap_p, name: str) -> str:
+    return xkbcommon.xkb_keymap_mod_get_index(keymap, name.encode("utf-8"))
+
+
+def xkb_keymap_mod_get_name(keymap: xkb_keymap_p, mod: int) -> str:
+    return xkbcommon.xkb_keymap_mod_get_name(keymap, mod).decode("utf-8")
+
+
+def xkb_keymap_get_mod_mask_from_names(
+    keymap: xkb_keymap_p, names: Iterable[str]
+) -> ModifierMask:
+    mods = filter(
+        lambda m: m != XKB_MOD_INVALID,
+        (xkb_keymap_mod_get_index(keymap, name) for name in names),
+    )
+    return reduce(lambda acc, m: acc | ModifierMask(1 << m), mods, ModifierMask(0))
+
+
+def xkb_keymap_get_usual_mod_mapping(
+    keymap: xkb_keymap_p, mask: ModifierMask
+) -> ModifierMask:
+    mods = (
+        "Alt" if Mod1 in mask else "",
+        "Meta" if Mod1 in mask else "",
+        "NumLock" if Mod2 in mask else "",
+        "LevelFive" if Mod3 in mask else "",
+        "Super" if Mod4 in mask else "",
+        "LevelThree" if Mod5 in mask else "",
+    )
+    return mask | xkb_keymap_get_mod_mask_from_names(keymap, mods)
+
+
 def xkb_keymap_led_get_name(keymap: xkb_keymap_p, led: int) -> str:
     return xkbcommon.xkb_keymap_led_get_name(keymap, led).decode("utf-8")
 
 
+def xkb_keysym_from_name(name: str) -> int:
+    return xkbcommon.xkb_keysym_from_name(name.encode("utf-8"), XKB_KEYSYM_NO_FLAGS)
+
+
 def xkb_keysym_get_name(keysym: int) -> str:
     buf_len = 90
     buf = create_string_buffer(buf_len)
@@ -381,6 +511,18 @@ def process_key_event(
 ###############################################################################
 
 
+@dataclass
+class KeyLevel:
+    """
+    Output of a key at a specific group and level.
+    """
+
+    keycode: int
+    group: int
+    level: int
+    keysyms: tuple[int, ...]
+
+
 class ForeignKeymap:
     """
     Context manager to ensure proper handling of foreign `xkb_keymap` object.
@@ -423,6 +565,52 @@ class ForeignKeymap:
     def __exit__(self, exc_type, exc_val, exc_tb):
         unref_keymap(self._keymap)
 
+    def check(self) -> bool:
+        """
+        Test if keymap compiles with corresponding RMLVO config.
+        """
+        try:
+            with self:
+                return bool(self._keymap)
+        except ValueError:
+            return False
+
+    def as_string(self) -> str:
+        """
+        Export a keymap as a string, or return an empty string on error.
+        """
+        try:
+            with self as keymap:
+                if not bool(keymap):
+                    return ""
+                return keymap_as_string(keymap)
+        except ValueError:
+            return ""
+
+    @classmethod
+    def __iter_key(
+        cls, result: list[KeyLevel], keymap: xkb_keymap_p, keycode: int, data: c_void_p
+    ):
+        """
+        Given a keymap and a keycode, iterate over its groups and levels and
+        append its outputs to a list of results.
+        """
+        max_groups = xkb_keymap_num_layouts_for_key(keymap, keycode)
+        for g in range(0, max_groups):
+            max_levels = xkb_keymap_num_levels_for_key(keymap, keycode, g)
+            for l in range(0, max_levels):
+                keysyms = xkb_keymap_key_get_syms_by_level(keymap, keycode, g, l)
+                result.append(KeyLevel(keycode, g, l, keysyms))
+
+    @classmethod
+    def get_keys_levels(cls, keymap: xkb_keymap_p) -> list[KeyLevel]:
+        """
+        Given a keymap, get the list of outputs for each key, group and level.
+        """
+        result: list[KeyLevel] = []
+        xkb_keymap_key_for_each(keymap, partial(cls.__iter_key, result), c_void_p())
+        return result
+
 
 class State:
     """
@@ -455,3 +643,20 @@ class ForeignState:
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         unref_state(self._state)
+
+
+def has_vmod_queries(xkb_base: Path) -> bool:
+    """
+    Check if the used xkbcommon support querying virtual modifiers
+    """
+    with ForeignKeymap(xkb_base) as keymap:
+        mod = xkb_keymap_mod_get_index(keymap, "Alt")
+        Alt = ModifierMask((1 << mod) if mod != XKB_MOD_INVALID else 0)
+        assert Alt.value != 0
+        with ForeignState(keymap) as state:
+            # This is very indirect: if vmods are supported in queries, then
+            # we expect Result.active_mods to contain rmods and relevant vmods;
+            # else we will get only rmods.
+            r = state.process_key_event("LALT", xkb_key_direction.XKB_KEY_DOWN)
+            r = state.process_key_event("AD01", xkb_key_direction.XKB_KEY_DOWN)
+            return Alt in r.active_mods
